@@ -9,7 +9,6 @@ import json
 import logging
 import uuid
 
-
 import dsnparse
 from datatypes import (
     Datetime,
@@ -215,6 +214,13 @@ class Schema(object):
         """the magic fields for the schema, magic fields start with an underscore"""
         return {f:v for f, v in self.fields.items() if f.startswith('_')}
 
+    @property
+    def pk_fields(self):
+        pk_fields = {}
+        for pk_name in self.pk_names:
+            pk_fields[pk_name] = self.fields[pk_name]
+        return pk_fields
+
     @property
     def pk_name(self):
         """returns the primary key name for this schema"""
@@ -553,9 +559,9 @@ class Field(object, metaclass=FieldMeta):
             return int(val)
 
         @bar.jsonabler
-        def bar(self, val):
-            # convert val to something json safe
-            return val
+        def bar(self, name, val):
+            # convert val to something json safe, you can also change the name
+            return name, val
 
     NOTE -- all these get/set/delete methods are different than traditional python getters
     and setters because they always need to return a value and they always take in the
@@ -596,9 +602,9 @@ class Field(object, metaclass=FieldMeta):
                 print("iquery")
                 return v
 
-            def jsonable(self, orm, v):
+            def jsonable(self, orm, field_name, v):
                 print("jsonable")
-                return v
+                return field_name, v
 
     https://docs.python.org/2/howto/descriptor.html
     """
@@ -700,6 +706,9 @@ class Field(object, metaclass=FieldMeta):
 
     @property
     def interface_options(self):
+        """When the interface is doing stuff with the field it might call this
+        method to make sure it has all the options it needs to use the field
+        """
         if self.is_ref():
             options = self.schema.pk.options
             options.update(self.options)
@@ -711,29 +720,29 @@ class Field(object, metaclass=FieldMeta):
         """
         create a field
 
-        :param field_type: type, the python type of the field, so for a string you would pass str, integer: int,
-            boolean: bool, float: float, big int: long
+        :param field_type: type, the python type of the field, so for a string
+            you would pass str, integer: int, boolean: bool, float: float
         :param field_required: boolean, true if this field has to be there to insert
         :param field_options: dict, everything else in key: val notation. Current options:
-            size -- int -- the size you want the string to be, or the int to be
-            min_size -- int -- the minimum size
-            max_size -- int -- if you want a varchar, set this
-            unique -- boolean -- True to set a unique index on this field, this is just for convenience and is
+            - size: int -- the size you want the string to be, or the int to be
+            - min_size: int -- the minimum size
+            - max_size: int -- if you want a varchar, set this
+            - unique: bool -- True to set a unique index on this field, this is just for convenience and is
                 equal to self.set_index(field_name, [field_name], unique=True). this is a convenience option
                 to set a unique index on the field without having to add a separate index statement
-            ignore_case -- boolean -- True to ignore case if this field is used in indexes
-            default -- mixed -- defaults to None, can be anything the db can support
+            - ignore_case: bool -- True to ignore case if this field is used in indexes
+            - default: mixed -- defaults to None, can be anything the db can support
+            - jsonable_name: str, the name of the field when Orm.jsonable() is called
+            - jsonable_field: bool, if False then this field won't be part of
+                Orm.jsonable() output
+            - jsonable: str|callable|bool,
+                - str, will be set to jsonable_name
+                - callable, will be used as self.jsonable
+                - bool, will set jsonable_field to False
+            - empty: bool, (default is True), set to False if the value cannot be
+                empty when being sent to the db (empty is None, "", 0, False)
         :param **field_options_kwargs: dict, will be combined with field_options
         """
-        # we aren't guaranteed to have this field's name when the descriptor is
-        # created, so this will be the field name this descriptor will use to 
-        # set the value onto the orm
-        #self.orm_field_name = "_instance_{}_val".format(id(self))
-
-        # we keep a hash of the field's value when it was pulled from the
-        # interface (see .iget) so we know if the field has been modified
-        #self.orm_interface_hash = "_interface_{}_hash".format(id(self))
-
         field_options = utils.make_dict(field_options, field_options_kwargs)
 
         d = self.get_size(field_options)
@@ -747,6 +756,17 @@ class Field(object, metaclass=FieldMeta):
         if choices or not self.choices:
             self.choices = choices
 
+        if jsonable := field_options.pop("jsonable", None):
+            if isinstance(jsonable, str):
+                field_options.setdefault("jsonable_name", jsonable)
+
+            elif isinstance(jsonable, bool):
+                field_options.setdefault("jsonable_field", jsonable)
+
+            else:
+                field_options["jsonable"] = jsonable
+
+
         for k in list(field_options.keys()):
             if hasattr(self, k):
                 setattr(self, k, field_options.pop(k))
@@ -952,7 +972,7 @@ class Field(object, metaclass=FieldMeta):
 
             elif issubclass(field_type, pickle_types):
                 self.serializer = self.options.pop("serializer", "pickle")
-                self._interface_type = str
+                self._interface_type = bytes
 
             else:
                 schema = getattr(field_type, "schema", None)
@@ -962,7 +982,7 @@ class Field(object, metaclass=FieldMeta):
                 else:
                     # We have just some random class that isn't an Orm
                     self.serializer = self.options.pop("serializer", "pickle")
-                    self._interface_type = str
+                    self._interface_type = bytes
 
         elif isinstance(field_type, Schema):
             self._schema = field_type
@@ -1005,7 +1025,7 @@ class Field(object, metaclass=FieldMeta):
         :param val: mixed, the current value of the field
         :returns: mixed
         """
-        #pout.v("fget {}".format(self.name))
+        logger.debug(f"fget {orm.__class__.__name__}.{self.name}")
         if self.is_ref():
             # Foreign Keys get passed through their Field methods
             val = self.schema.pk.fget(orm, val)
@@ -1027,7 +1047,7 @@ class Field(object, metaclass=FieldMeta):
         :param val: mixed, the current value of the field
         :returns: mixed
         """
-        #pout.v("iget {}".format(self.name))
+        logger.debug(f"iget {orm.__class__.__name__}.{self.name}")
         if self.is_ref():
             # Foreign Keys get passed through their Field methods
             val = self.schema.pk.iget(orm, val)
@@ -1049,18 +1069,22 @@ class Field(object, metaclass=FieldMeta):
     def fset(self, orm, val):
         """This is called on Orm instantiation and any time field is set (eg Orm.foo = ...)
 
-        on Orm creation val will be None if the field wasn't pass to Orm.__init__ otherwise
+        on Orm creation val will be None if the field wasn't passed to Orm.__init__ otherwise
         it will be the value passed into Orm.__init__
 
         :param orm: Orm, the Orm instance the field is being set on
         :param val: mixed, the current value of the field
         :returns: mixed
         """
-        #pout.v("fset {}".format(self.name))
-        if val is not None and self.choices:
-            if val not in self.choices:
+        logger.debug(f"fset {orm.__class__.__name__}.{self.name}")
+        if val is not None:
+            if self.choices and val not in self.choices:
                 raise ValueError("Value {} not in {} value choices".format(val, self.name))
 
+            if regex := self.options.get("regex", ""):
+                if not re.search(regex, val):
+                    raise ValueError(f"regex failed for {orm.__class__.__name__}.{self.name}")
+
         if self.is_ref():
             # Foreign Keys get passed through their Field methods
             val = self.schema.pk.fset(orm, val)
@@ -1075,13 +1099,16 @@ class Field(object, metaclass=FieldMeta):
     def iset(self, orm, val):
         """Called anytime the field is being fetched to send to the interface
 
-        think of this as when the interface is going to get the field value
+        think of this as when the interface is going to get the field value or
+        when the field is being sent to the db. Alot of the value checks like
+        required and empty are in Orm.to_interface(), this just returns the value
+        and nothing else
 
         :param orm: Orm
-        :param val: mixed, the current value of the field
-        :returns: mixed
+        :param val: Any, the current value of the field
+        :returns: Any
         """
-        #pout.v("iset {}".format(self.name))
+        logger.debug(f"iset {orm.__class__.__name__}.{self.name}")
         if self.is_ref():
             # Foreign Keys get passed through their Field methods
             val = self.schema.pk.iset(orm, val)
@@ -1098,6 +1125,7 @@ class Field(object, metaclass=FieldMeta):
         return self
 
     def fdel(self, orm, val):
+        logger.debug(f"fdel {orm.__class__.__name__}.{self.name}")
         orm.__dict__.pop(self.orm_interface_hash, None)
         return None
 
@@ -1113,6 +1141,7 @@ class Field(object, metaclass=FieldMeta):
         :param val: mixed, the current value of the field
         :returns: mixed
         """
+        logger.debug(f"idel {orm.__class__.__name__}.{self.name}")
         orm.__dict__.pop(self.orm_interface_hash, None)
         return None if self.is_pk() else val
 
@@ -1179,21 +1208,37 @@ class Field(object, metaclass=FieldMeta):
         self.iquery = v
         return self
 
-    def jsonable(self, orm, val):
-        if self.is_ref():
-            # Foreign Keys get passed through their Field methods
-            val = self.schema.pk.jsonable(orm, val)
+    def jsonable(self, orm, name, val):
+        """This is called in Orm.jsonable() to set the field name nad value
+
+        :param orm: Orm, the instance currently calling jsonable
+        :param name: str, the Orm field name that will be used if the jsonable_name
+            option isn't set
+        :param val: Any, the field's value
+        :returns: tuple[str, Any], (name, val) where the name will be the jsonable
+            field name and the value will be the jsonable value
+        """
+        if self.options.get("jsonable_field", True):
+            logger.debug(f"jsonable {orm.__class__.__name__}.{self.name} -> {name}")
+
+            if self.is_ref():
+                # Foreign Keys get passed through their Field methods
+                val = self.schema.pk.jsonable(orm, val)
+
+            else:
+                if val is None:
+                    val = self.fdefault(orm, val)
+
+                if val is not None:
+                    format_str = ""
+                    if isinstance(val, (datetime.datetime, datetime.date)):
+                        val = Datetime(val).isoformat()
+
+            return self.options.get("jsonable_name", name), val
 
         else:
-            if val is None:
-                val = self.fdefault(orm, val)
-
-            if val is not None:
-                format_str = ""
-                if isinstance(val, (datetime.datetime, datetime.date)):
-                    val = Datetime(val).isoformat()
-
-        return val
+            logger.debug(f"jsonable ignored for {orm.__class__.__name__}.{self.name}")
+            return "", None
 
     def jsonabler(self, v):
         """Decorator for the method called for a field when an Orm's .jsonable method
@@ -1232,7 +1277,8 @@ class Field(object, metaclass=FieldMeta):
         if val is None: return val
 
         if self.serializer == "pickle":
-            return base64.b64encode(pickle.dumps(val, pickle.HIGHEST_PROTOCOL))
+            #return base64.b64encode(pickle.dumps(val, pickle.HIGHEST_PROTOCOL))
+            return pickle.dumps(val, pickle.HIGHEST_PROTOCOL)
 
         elif self.serializer == "json":
             return json.dumps(val)
@@ -1244,7 +1290,8 @@ class Field(object, metaclass=FieldMeta):
         if val is None: return val
 
         if self.serializer == "pickle":
-            return pickle.loads(base64.b64decode(val))
+            #return pickle.loads(base64.b64decode(val))
+            return pickle.loads(val)
 
         elif self.serializer == "json":
             return json.loads(val)
@@ -1256,6 +1303,7 @@ class Field(object, metaclass=FieldMeta):
         """return the raw value that this property is holding internally for the orm instance"""
         try:
             val = orm.__dict__[self.orm_field_name]
+
         except KeyError as e:
             #raise AttributeError(str(e))
             val = None
