@@ -7,6 +7,7 @@ http://wiki.postgresql.org/wiki/Using_psycopg2_with_PostgreSQL
 http://pythonhosted.org/psycopg2/
 """
 from __future__ import unicode_literals, division, print_function, absolute_import
+import re
 import os
 import sys
 import decimal
@@ -30,6 +31,7 @@ from ...exception import (
     FieldError,
     UniqueError,
     CloseError,
+    PlaceholderError,
 )
 
 
@@ -109,40 +111,38 @@ class PostgreSQLConnection(SQLConnection, psycopg2.extensions.connection):
         # https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.register_adapter
         psycopg2.extensions.register_adapter(dict, DictType.adapt)
 
-        # http://initd.org/psycopg/docs/connection.html#connection.set_client_encoding
-        # https://www.postgresql.org/docs/current/static/multibyte.html
-        # > The default is the encoding defined by the database
-        # Not sure we want to override db encoding which is probably why I didn't
-        # set this previously
-        #self.set_client_encoding("UTF8")
-
-        #self.initialize(logger)
-
 
 class PostgreSQL(SQLInterface):
     """
     https://www.psycopg.org/docs/
     https://www.psycopg.org/docs/usage.html
     """
-    val_placeholder = '%s'
+    LIMIT_NONE = "ALL"
 
     _connection_pool = None
 
-    def _connect(self, connection_config):
-        database = connection_config.database
-        username = connection_config.username
-        password = connection_config.password
-        host = connection_config.host
-        port = connection_config.port
-        if not port: port = 5432
+    @classmethod
+    def configure(cls, config):
+        config = super().configure(config)
+        port = config.port
+        if not port:
+            config.port = 5432
+        return config
 
+    def get_paramstyle(self):
+        """
+        https://www.psycopg.org/docs/module.html#psycopg2.paramstyle
+        """
+        return psycopg2.paramstyle
+
+    def _connect(self, config):
         # set pool_key to None to use the pool as an actual pool, right now we've
         # turned it back into basically a sync interface since I don't need the gevent
         # stuff right now and I want to switch to using the native async interface
-        self.connection_config.options.setdefault("pool_key", "sync_interface")
-        minconn = int(connection_config.options.get('pool_minconn', 1))
-        maxconn = int(connection_config.options.get('pool_maxconn', 1))
-        pool_class_name = connection_config.options.get(
+        self.config.options.setdefault("pool_key", "sync_interface")
+        minconn = int(config.options.get('pool_minconn', 1))
+        maxconn = int(config.options.get('pool_maxconn', 1))
+        pool_class_name = config.options.get(
             'pool_class',
             'psycopg2.pool.SimpleConnectionPool'
         )
@@ -156,11 +156,11 @@ class PostgreSQL(SQLInterface):
         self._connection_pool = pool_class(
             minconn,
             maxconn,
-            dbname=database,
-            user=username,
-            password=password,
-            host=host,
-            port=port,
+            dbname=config.database,
+            user=config.username,
+            password=config.password,
+            host=config.host,
+            port=config.port,
             cursor_factory=psycopg2.extras.RealDictCursor,
             #cursor_factory=LoggingCursor,
             connection_factory=PostgreSQLConnection,
@@ -168,20 +168,20 @@ class PostgreSQL(SQLInterface):
 
     def _get_connection(self):
         connection = self._connection_pool.getconn(
-            key=self.connection_config.options["pool_key"]
+            key=self.config.options["pool_key"]
         )
 
         connection_id = id(connection)
 
         # change the connection readonly status if they don't match
-        if bool(connection.readonly) != bool(self.connection_config.readonly):
+        if bool(connection.readonly) != bool(self.config.readonly):
             # https://www.psycopg.org/docs/connection.html#connection.readonly
             self.log_warning([
                 f"Changing connection {connection_id:02x}",
-                f"to readonly={self.connection_config.readonly}",
+                f"to readonly={self.config.readonly}",
                 f"from readonly={connection.readonly}",
             ])
-            connection.readonly = self.connection_config.readonly
+            connection.readonly = self.config.readonly
 
         return connection
 
@@ -193,7 +193,7 @@ class PostgreSQL(SQLInterface):
                     self.log_warning(f"Discarding pool connection {id(connection):02x} because it is closed")
                     self._connection_pool.putconn(
                         connection,
-                        key=self.connection_config.options["pool_key"],
+                        key=self.config.options["pool_key"],
                         close=True,
                     )
 
@@ -216,7 +216,7 @@ class PostgreSQL(SQLInterface):
 
     def _get_tables(self, table_name, **kwargs):
         query_str = 'SELECT tablename FROM pg_tables WHERE tableowner = %s'
-        query_args = [self.connection_config.username]
+        query_args = [self.config.username]
 
         if table_name:
             query_str += ' AND tablename = %s'
@@ -227,7 +227,7 @@ class PostgreSQL(SQLInterface):
             query_str += ' AND schemaname = %s'
             query_args.append(schema_name)
 
-        ret = self._query(query_str, *query_args, **kwargs)
+        ret = self._raw(query_str, *query_args, **kwargs)
         # http://www.postgresql.org/message-id/CA+mi_8Y6UXtAmYKKBZAHBoY7F6giuT5WfE0wi3hR44XXYDsXzg@mail.gmail.com
         return [r['tablename'] for r in ret]
 
@@ -235,13 +235,13 @@ class PostgreSQL(SQLInterface):
         """
         https://www.postgresql.org/docs/current/sql-droptable.html
         """
-        query_str = 'DROP TABLE IF EXISTS {} CASCADE'.format(self._normalize_table_name(schema))
-        ret = self._query(query_str, ignore_result=True, **kwargs)
+        query_str = 'DROP TABLE IF EXISTS {} CASCADE'.format(self.render_table_name_sql(schema))
+        self._raw(query_str, ignore_result=True, **kwargs)
 
     def _get_fields(self, table_name, **kwargs):
         """return all the fields for the given schema"""
         ret = {}
-        query_args = ['f', self._normalize_table_name(table_name)]
+        query_args = ['f', self.render_table_name_sql(table_name)]
 
         # I had to brush up on my join knowledge while writing this query
         # https://en.wikipedia.org/wiki/Join_(SQL)
@@ -272,16 +272,16 @@ class PostgreSQL(SQLInterface):
             'LEFT JOIN pg_index i ON a.attrelid = i.indrelid',
             '  AND a.attnum = any(i.indkey)',
             'LEFT JOIN pg_constraint s ON a.attrelid = s.conrelid',
-            '  AND s.contype = {} AND a.attnum = any(s.conkey)'.format(self.val_placeholder),
+            '  AND s.contype = {} AND a.attnum = any(s.conkey)'.format(self.PLACEHOLDER),
             'LEFT JOIN pg_class c ON s.confrelid = c.oid',
             'WHERE',
-            '  a.attrelid = {}::regclass'.format(self.val_placeholder),
+            '  a.attrelid = {}::regclass'.format(self.PLACEHOLDER),
             '  AND a.attisdropped = False',
             '  AND a.attnum > 0',
             'ORDER BY a.attnum ASC',
         ]
         query_str = "\n".join(query_str)
-        fields = self._query(query_str, *query_args, **kwargs)
+        fields = self._raw(query_str, *query_args, **kwargs)
 
         pg_types = {
             "float4": float,
@@ -297,8 +297,9 @@ class PostgreSQL(SQLInterface):
             "varchar": str,
             "bool": bool,
             "date": datetime.date,
-            "blob": bytearray,
-            "jsonb": bytearray,
+            "blob": bytes,
+            "bytea": bytes,
+            "jsonb": bytes,
         }
 
         # the rows we can set: field_type, name, field_required, min_size, max_size,
@@ -338,7 +339,7 @@ class PostgreSQL(SQLInterface):
         ]
         query_str = "\n".join(query_str)
 
-        indexes = self._query(query_str, 'r', str(schema), **kwargs)
+        indexes = self._raw(query_str, 'r', str(schema), **kwargs)
 
         # massage the data into more readable {index_name: fields} format
         for idict in indexes:
@@ -350,16 +351,7 @@ class PostgreSQL(SQLInterface):
 
         return ret
 
-    def _normalize_field_SQL(self, schema, field_name, symbol):
-        format_field_name = self._normalize_name(field_name)
-        format_val_str = self.val_placeholder
-
-        if 'LIKE' in symbol:
-            format_field_name += '::text'
-
-        return format_field_name, format_val_str
-
-    def _normalize_sort_SQL(self, field_name, field_vals, sort_dir_str):
+    def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
         # this solution is based off:
         # http://postgresql.1045698.n5.nabble.com/ORDER-BY-FIELD-feature-td1901324.html
         # see also: https://gist.github.com/cpjolicoeur/3590737
@@ -367,12 +359,12 @@ class PostgreSQL(SQLInterface):
         query_args = []
         for v in reversed(field_vals):
             query_sort_str.append('  {} = {} {}'.format(
-                self._normalize_name(field_name), self.val_placeholder, sort_dir_str))
+                self.render_field_name_sql(field_name), self.PLACEHOLDER, sort_dir_str))
             query_args.append(v)
 
         return ',\n'.join(query_sort_str), query_args
 
-    def _normalize_date_SQL(self, field_name, field_kwargs, symbol):
+    def render_date_field_sql(self, field_name, field_kwargs, symbol):
         """
         allow extracting information from date
 
@@ -397,25 +389,10 @@ class PostgreSQL(SQLInterface):
         }
 
         for k, v in field_kwargs.items():
-            fstrs.append([k_opts[k].format(self._normalize_name(field_name)), self.val_placeholder, v])
+            fstrs.append([k_opts[k].format(self.render_field_name_sql(field_name)), self.PLACEHOLDER, v])
 
         return fstrs
 
-    def _normalize_bounds_SQL(self, bounds, sql_options):
-        offset = bounds.offset
-        if sql_options.get('one_query', False):
-            limit = 1
-
-        else:
-            limit, offset = bounds.get()
-            if not bounds.has_limit():
-                limit = "ALL"
-
-        return 'LIMIT {} OFFSET {}'.format(
-            limit,
-            offset
-        )
-
     def render_datatype_int_sql(self, field_name, field, **kwargs):
         if field.is_pk():
             field_type = 'BIGSERIAL PRIMARY KEY' # INT8
@@ -495,6 +472,16 @@ class PostgreSQL(SQLInterface):
 
         return field_type
 
+    def render_datatype_bytes_sql(self, field_name, field, **kwargs):
+        """Why Postgres? Why?
+
+        https://www.postgresql.org/docs/current/datatype-binary.html
+            The SQL standard defines a different binary string type, called BLOB
+            or BINARY LARGE OBJECT. The input format is different from bytea, but
+            the provided functions and operators are mostly the same.
+        """
+        return 'BYTEA'
+
     def render_datatype_uuid_sql(self, field_name, field, **kwargs):
         # https://www.postgresql.org/docs/current/datatype-uuid.html
         # https://www.postgresql.org/docs/current/functions-uuid.html
@@ -504,6 +491,10 @@ class PostgreSQL(SQLInterface):
         return field_type
 
     def create_error(self, e, **kwargs):
+        """
+        https://www.psycopg.org/docs/module.html#exceptions
+        """
+        kwargs.setdefault("error_module", psycopg2)
         if isinstance(e, psycopg2.ProgrammingError):
             e_msg = String(e)
             if "does not exist" in e_msg:
@@ -516,6 +507,22 @@ class PostgreSQL(SQLInterface):
                     #'relation "<TABLE-NAME>" does not exit'
                     e = TableError(e)
 
+            elif "can't adapt type" in e_msg and "error_args" in kwargs:
+                error_args = kwargs["error_args"]
+                ms = re.search(r"type\s'(\S+)'", e_msg)
+                type_name = ms.group(1)
+                for index, value in enumerate(error_args[1]):
+                    if type_name in str(type(value)):
+                        e = PlaceholderError(
+                            e,
+                            *error_args,
+                            message=f"Placeholder {index} of query has unexpected type {type(value)}",
+                        )
+                        break
+
+            else:
+                e = super().create_error(e, **kwargs)
+
         elif isinstance(e, psycopg2.errors.AdminShutdown):
             e = CloseError(e)
 
@@ -525,6 +532,9 @@ class PostgreSQL(SQLInterface):
         elif isinstance(e, psycopg2.IntegrityError):
             e = UniqueError(e)
 
+        elif isinstance(e, IndexError) and "error_args" in kwargs:
+            e = PlaceholderError(e, *kwargs.get("error_args", []))
+
         else:
             e = super().create_error(e, **kwargs)
 
