@@ -5,12 +5,14 @@ import datetime
 import decimal
 import logging
 import uuid
+from functools import cached_property
 
 from ..query import Query
 from ..exception import (
     TableError,
     FieldError,
     UniqueError,
+    PlaceholderError,
 )
 
 from ..compat import *
@@ -32,7 +34,7 @@ class SQLConnection(Connection):
 
     def _transaction_started(self, name):
         # http://www.postgresql.org/docs/9.2/static/sql-savepoint.html
-        self._execute("SAVEPOINT {}".format(self.interface._normalize_name(name)))
+        self._execute("SAVEPOINT {}".format(self.interface.render_field_name_sql(name)))
 
     def _transaction_stop(self):
         """
@@ -42,14 +44,14 @@ class SQLConnection(Connection):
         self._execute("COMMIT")
 
     def _transaction_stopping(self, name):
-        self._execute("RELEASE {}".format(self.interface._normalize_name(name)))
+        self._execute("RELEASE {}".format(self.interface.render_field_name_sql(name)))
 
     def _transaction_fail(self):
         self._execute("ROLLBACK")
 
     def _transaction_failing(self, name):
         # http://www.postgresql.org/docs/9.2/static/sql-rollback-to.html
-        self._execute("ROLLBACK TO SAVEPOINT {}".format(self.interface._normalize_name(name)))
+        self._execute("ROLLBACK TO SAVEPOINT {}".format(self.interface.render_field_name_sql(name)))
 
     def _execute(self, query_str):
         self.log_info(f"0x{id(self):02x} - {query_str}")
@@ -62,22 +64,38 @@ class SQLInterfaceABC(Interface):
     """SQL database interfaces should extend SQLInterface and implement all these
     methods in this class and all the methods in InterfaceABC"""
     @property
-    def val_placeholder(self):
+    def LIMIT_NONE(self):
+        """When an offset is set but not a limit, this is the value that will be
+        put into the LIMIT part of the query
+
+        :returns: str|int|None
+        """
         raise NotImplementedError("this property should be set in any children class")
 
-    def _normalize_date_SQL(self, field_name, field_kwargs, symbol):
+    def get_paramstyle(self):
+        """Returns the paramstyle that is used by self.PLACEHOLDER to decide what
+        val placeholder to use when building queries. This would also be the value
+        to use when calling .raw().
+
+        The dbapi 2.0 spec requires this to be set
+
+        https://peps.python.org/pep-0249/#paramstyle
+
+        :returns: str, something like "qmark" (for ? placeholders) or "pyformat"
+            (for %s placeholders)
+        """
         raise NotImplemented()
 
-    def _normalize_field_SQL(self, schema, field_name, symbol):
-        return self._normalize_name(field_name), self.val_placeholder
+    def render_date_field_sql(self, field_name, field_kwargs, symbol):
+        raise NotImplemented()
 
-    def _normalize_sort_SQL(self, field_name, field_vals, sort_dir_str):
+    def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
         """normalize the sort string
 
         return -- tuple -- field_sort_str, field_sort_args"""
         raise NotImplemented()
 
-    def _normalize_bounds_SQL(self, bounds):
+    def render_nolimit_sql(limit, **kwargs):
         raise NotImplemented()
 
     def render_datatype_datetime_sql(self, field_name, field, **kwargs):
@@ -92,6 +110,40 @@ class SQLInterfaceABC(Interface):
 
 class SQLInterface(SQLInterfaceABC):
     """Generic base class for all SQL derived interfaces"""
+    @cached_property
+    def PLACEHOLDER(self):
+        """What placeholder value this interface uses when building queries.
+
+        This uses .get_paramstyle() to decide what placeholder value to use
+
+        https://www.psycopg.org/docs/usage.html#passing-parameters-to-sql-queries
+
+        NOTE -- It looks like both SQLite and Postgres support "named" so if you
+            want to use .raw() queries that will work in both interface I would use
+            named parameters
+
+        :returns: str, usually something like "?" or "%s"
+        """
+        paramstyle = self.get_paramstyle()
+
+        if paramstyle == "qmark":
+            return "?"
+
+        elif paramstyle == "format":
+            return "%s"
+
+        elif paramstyle == "pyformat":
+            return "%s"
+
+        elif paramstyle == "numeric":
+            raise NotImplementedError("These are :1 :2 :3")
+
+        elif paramstyle == "named":
+            raise NotImplementedError("These are :name")
+
+        else:
+            raise NotImplementedError(f"Unknown paramstyle {paramstyle}")
+
     def _set_table(self, schema, **kwargs):
         """
         http://sqlite.org/lang_createtable.html
@@ -100,7 +152,7 @@ class SQLInterface(SQLInterfaceABC):
         http://pythonhosted.org/psycopg2/usage.html#adaptation-of-python-values-to-sql-types
         """
         query_str = []
-        query_str.append("CREATE TABLE IF NOT EXISTS {} (".format(self._normalize_table_name(schema)))
+        query_str.append("CREATE TABLE IF NOT EXISTS {} (".format(self.render_table_name_sql(schema)))
 
         query_fields = []
         for field_name, field in schema.fields.items():
@@ -109,7 +161,7 @@ class SQLInterface(SQLInterfaceABC):
         query_str.append(",\n".join(query_fields))
         query_str.append(')')
         query_str = "\n".join(query_str)
-        self._query(query_str, ignore_result=True, **kwargs)
+        self._raw(query_str, ignore_result=True, **kwargs)
 
     def _set_index(self, schema, name, field_names, **kwargs):
         """
@@ -122,12 +174,12 @@ class SQLInterface(SQLInterfaceABC):
         """
         query_str = 'CREATE {}INDEX IF NOT EXISTS {} ON {} ({})'.format(
             'UNIQUE ' if kwargs.get('unique', False) else '',
-            self._normalize_name(f"{schema}_{name}"),
-            self._normalize_table_name(schema),
-            ', '.join(map(self._normalize_name, field_names))
+            self.render_field_name_sql(f"{schema}_{name}"),
+            self.render_table_name_sql(schema),
+            ', '.join(map(self.render_field_name_sql, field_names))
         )
 
-        return self._query(query_str, ignore_result=True, **kwargs)
+        return self._raw(query_str, ignore_result=True, **kwargs)
 
     def _insert(self, schema, fields, **kwargs):
         pk_names = schema.pk_names
@@ -140,7 +192,7 @@ class SQLInterface(SQLInterfaceABC):
             **kwargs,
         )
 
-        r = self._query(query_str, *query_args, **kwargs)
+        r = self._raw(query_str, *query_args, **kwargs)
         if r and pk_names:
             if len(pk_names) > 1:
                 r = r[0]
@@ -156,7 +208,7 @@ class SQLInterface(SQLInterfaceABC):
             **kwargs,
         )
 
-        return self._query(query_str, *query_args, count_result=True, **kwargs)
+        return self._raw(query_str, *query_args, count_result=True, **kwargs)
 
     def _upsert(self, schema, insert_fields, update_fields, conflict_field_names, **kwargs):
         """
@@ -192,17 +244,17 @@ class SQLInterface(SQLInterfaceABC):
 
         query_str = '{} ON CONFLICT({}) DO UPDATE {}'.format(
             insert_sql,
-            ', '.join(map(self._normalize_name, conflict_field_names)),
+            ', '.join(map(self.render_field_name_sql, conflict_field_names)),
             update_sql,
         )
 
         returning_field_names = schema.pk_names
         if returning_field_names:
             # https://www.sqlite.org/lang_returning.html
-            query_str += ' RETURNING {}'.format(', '.join(map(self._normalize_name, returning_field_names)))
+            query_str += ' RETURNING {}'.format(', '.join(map(self.render_field_name_sql, returning_field_names)))
             query_args = insert_args + update_args
 
-        r = self._query(query_str, *query_args, **kwargs)
+        r = self._raw(query_str, *query_args, **kwargs)
         if r and returning_field_names:
             if len(returning_field_names) > 1:
                 r = r[0]
@@ -211,16 +263,16 @@ class SQLInterface(SQLInterfaceABC):
         return r
 
     def _delete(self, schema, query, **kwargs):
-        where_query_str, query_args = self.get_SQL(schema, query, only_where_clause=True)
+        where_query_str, query_args = self.render_sql(schema, query, only_where_clause=True)
         query_str = []
         query_str.append('DELETE FROM')
-        query_str.append('  {}'.format(self._normalize_table_name(schema)))
+        query_str.append('  {}'.format(self.render_table_name_sql(schema)))
         query_str.append(where_query_str)
         query_str = "\n".join(query_str)
-        ret = self._query(query_str, *query_args, count_result=True, **kwargs)
+        ret = self._raw(query_str, *query_args, count_result=True, **kwargs)
         return ret
 
-    def _query(self, query_str, *query_args, **kwargs):
+    def _raw(self, query_str, *query_args, **kwargs):
         """
         **kwargs -- dict
             ignore_result -- boolean -- true to not attempt to fetch results
@@ -232,17 +284,35 @@ class SQLInterface(SQLInterfaceABC):
         #connection = kwargs.get('connection', None)
         with self.connection(**kwargs) as connection:
             cur = connection.cursor()
+
             ignore_result = kwargs.get('ignore_result', False)
             count_result = kwargs.get('count_result', False)
             one_result = kwargs.get('fetchone', kwargs.get('one_result', False))
             cursor_result = kwargs.get('cursor_result', False)
 
+            if cursor_result:
+                # https://stackoverflow.com/a/125140
+                cur.arraysize = kwargs.get("arraysize", 500)
+
             if query_args:
                 self.log_for(
                     debug=(["0x{:02x} - {}\n{}", id(connection), query_str, query_args],),
                     info=([f"0x{id(connection):02x} - {query_str}"],)
                 )
-                cur.execute(query_str, query_args)
+
+                try:
+                    cur.execute(query_str, query_args)
+
+                except Exception as e:
+                    self.raise_error(
+                        e,
+                        error_args=[
+                            query_str,
+                            query_args,
+                            self.PLACEHOLDER,
+                        ]
+                    )
+
             else:
                 self.log_info(f"0x{id(connection):02x} - {query_str}")
                 cur.execute(query_str)
@@ -270,17 +340,16 @@ class SQLInterface(SQLInterfaceABC):
 
             return ret
 
-    def _get_one(self, schema, query, **kwargs):
-        query_str, query_args = self.get_SQL(schema, query, one_query=True)
-        return self._query(query_str, *query_args, fetchone=True, **kwargs)
-
     def _get(self, schema, query, **kwargs):
-        query_str, query_args = self.get_SQL(schema, query)
-        return self._query(query_str, *query_args, **kwargs)
+        """
+        https://www.sqlite.org/lang_select.html
+        """
+        query_str, query_args = self.render_sql(schema, query)
+        return self._raw(query_str, *query_args, **kwargs)
 
     def _count(self, schema, query, **kwargs):
-        query_str, query_args = self.get_SQL(schema, query, count_query=True)
-        ret = self._query(query_str, *query_args, **kwargs)
+        query_str, query_args = self.render_sql(schema, query, count_query=True)
+        ret = self._raw(query_str, *query_args, **kwargs)
         if ret:
             ret = int(ret[0]['ct'])
         else:
@@ -304,11 +373,11 @@ class SQLInterface(SQLInterfaceABC):
                 else:
                     query_str = []
                     query_str.append('ALTER TABLE')
-                    query_str.append('  {}'.format(self._normalize_table_name(schema)))
+                    query_str.append('  {}'.format(self.render_table_name_sql(schema)))
                     query_str.append('ADD COLUMN')
                     query_str.append('  {}'.format(self.render_datatype_sql(field_name, field)))
                     query_str = "\n".join(query_str)
-                    self._query(query_str, ignore_result=True, **kwargs)
+                    self._raw(query_str, ignore_result=True, **kwargs)
 
         return True
 
@@ -338,7 +407,7 @@ class SQLInterface(SQLInterfaceABC):
         self.set_table(schema, **kwargs)
         return True
 
-    def _normalize_val_SQL(self, schema, symbol_map, field):
+    def render_field_sql(self, schema, symbol_map, field):
         format_str = ''
         format_args = []
         symbol = symbol_map['symbol']
@@ -351,7 +420,7 @@ class SQLInterface(SQLInterfaceABC):
             # kwargs take precedence because None is a perfectly valid field_val
             f = schema.fields[field_name]
             if issubclass(f.type, (datetime.datetime, datetime.date)):
-                format_strs = self._normalize_date_SQL(field_name, field_kwargs, symbol)
+                format_strs = self.render_date_field_sql(field_name, field_kwargs, symbol)
                 for fname, fvstr, farg in format_strs:
                     if format_str:
                         format_str += ' AND '
@@ -380,7 +449,9 @@ class SQLInterface(SQLInterfaceABC):
         else:
             if is_list and not isinstance(field_val, Query):
                 field_val = make_list(field_val) if field_val else []
-                field_name, format_val_str = self._normalize_field_SQL(schema, field_name, symbol)
+                field_name = self.render_field_name_sql(field_name)
+                format_val_str = self.PLACEHOLDER
+
                 if field_val:
                     format_str = '{} {} ({})'.format(
                         field_name,
@@ -412,23 +483,20 @@ class SQLInterface(SQLInterfaceABC):
                         raise ValueError("Unsure what to do here")
 
             else:
+                field_name = self.render_field_name_sql(field_name)
+                format_val_str = self.PLACEHOLDER
+
                 # special handling for NULL
                 if field_val is None:
                     symbol = symbol_map['none_symbol']
 
-                field_name, format_val_str = self._normalize_field_SQL(
-                    schema,
-                    field_name,
-                    symbol
-                )
-
                 if isinstance(field_val, Query):
                     subquery_schema = field_val.schema
                     if not subquery_schema:
                         raise ValueError("{} subquery has no schema".format(field_name))
 
-                    subquery_sql, subquery_args = self.get_SQL(
-                        field_val.schema,
+                    subquery_sql, subquery_args = self.render_sql(
+                        subquery_schema,
                         field_val
                     )
 
@@ -443,16 +511,16 @@ class SQLInterface(SQLInterfaceABC):
                     format_str = '{} {} {}'.format(
                         field_name,
                         symbol,
-                        format_val_str
+                        format_val_str,
                     )
                     format_args.append(field_val)
 
         return format_str, format_args
 
-    def _normalize_table_name(self, schema):
-        return self._normalize_name(schema)
+    def render_table_name_sql(self, schema):
+        return self.render_field_name_sql(schema)
 
-    def _normalize_name(self, name):
+    def render_field_name_sql(self, name):
         """normalize a non value name for the query
 
         https://blog.christosoft.de/2012/10/sqlite-escaping-table-acolumn-names/
@@ -462,18 +530,85 @@ class SQLInterface(SQLInterfaceABC):
         """
         return '"{}"'.format(name)
 
-    def get_SQL(self, schema, query, **sql_options):
-        """
-        convert the query instance into SQL
+    def render_select_sql(self, schema, query, **kwargs):
+        query_str = []
+        query_args = []
 
-        this is the glue method that translates the generic Query() instance to
-        the SQL specific query, this is where the magic happens
+        only_where_clause = kwargs.get('only_where_clause', False)
+        if not only_where_clause:
+            query_str.append('SELECT')
+            is_count_query = kwargs.get('count_query', False)
+
+            if query.compounds:
+                select_fields_str = "*"
+
+            else:
+                select_fields = query.fields_select
+                if select_fields:
+                    distinct_fields = select_fields.options.get(
+                        "distinct",
+                        select_fields.options.get("unique", False)
+                    )
+                    distinct = "DISTINCT " if distinct_fields else ""
+                    select_fields_str = distinct + ", ".join(
+                        (self.render_field_name_sql(f.name) for f in select_fields)
+                    )
+
+                else:
+                    if is_count_query or select_fields.options.get("all", False):
+                        select_fields_str = "*"
+
+                    else:
+                        select_fields_str = ", ".join(
+                            (self.render_field_name_sql(fname) for fname in schema.fields.keys())
+                        )
+
+            if is_count_query:
+                query_str.append('  count({}) as ct'.format(select_fields_str))
+
+            else:
+                query_str.append('  {}'.format(select_fields_str))
+
+            query_str.append('FROM')
+
+            if not query.compounds:
+                query_str.append("  {}".format(self.render_table_name_sql(schema)))
+
+        return query_str, query_args
+
+    def render_compound_sql(self, schema, query, **kwargs):
+        query_str = []
+        query_args = []
+
+        if query.compounds:
+            query_str.append("(")
+            for operator, queries in query.compounds:
+                for i, query in enumerate(queries):
+                    if i > 0:
+                        query_str.append(operator.upper())
+
+                    subquery_str, subquery_args = self.render_sql(
+                        query.schema,
+                        query,
+                    )
+
+                    if query.bounds:
+                        query_str.append("  (")
+                        query_str.append(subquery_str)
+                        query_str.append("  )")
+
+                    else:
+                        query_str.append(subquery_str)
+
+                    query_args.extend(subquery_args)
+
+            query_str.append(") I")
+        return query_str, query_args
+
+    def render_where_sql(self, schema, query, **kwargs):
+        query_str = []
+        query_args = []
 
-        **sql_options -- dict
-            count_query -- boolean -- true if this is a count query SELECT
-            only_where_clause -- boolean -- true to only return after WHERE ...
-        """
-        only_where_clause = sql_options.get('only_where_clause', False)
         symbol_map = {
             'in': {'symbol': 'IN', 'list': True},
             'nin': {'symbol': 'NOT IN', 'list': True},
@@ -489,75 +624,148 @@ class SQLInterface(SQLInterfaceABC):
             'nlike': {'symbol': 'NOT LIKE'},
         }
 
-        query_args = []
-        query_str = []
-
-        if not only_where_clause:
-            query_str.append('SELECT')
-            is_count_query = sql_options.get('count_query', False)
-            select_fields = query.fields_select
-            if select_fields:
-                distinct_fields = select_fields.options.get(
-                    "distinct",
-                    select_fields.options.get("unique", False)
-                )
-                distinct = "DISTINCT " if distinct_fields else ""
-                select_fields_str = distinct + ", ".join(
-                    (self._normalize_name(f.name) for f in select_fields)
-                )
-            else:
-                if is_count_query or select_fields.options.get("all", False):
-                    select_fields_str = "*"
-                else:
-                    select_fields_str = ", ".join(
-                        (self._normalize_name(fname) for fname in schema.fields.keys())
-                    )
-
-            if is_count_query:
-                query_str.append('  count({}) as ct'.format(select_fields_str))
-
-            else:
-                query_str.append('  {}'.format(select_fields_str))
-
-            query_str.append('FROM')
-            query_str.append("  {}".format(self._normalize_table_name(schema)))
-
         if query.fields_where:
             query_str.append('WHERE')
+            or_clause = False
 
             for i, field in enumerate(query.fields_where):
-                if i > 0: query_str.append('AND')
+                if i > 0:
+                    query_str.append('OR' if or_clause else 'AND')
 
                 field_str = ''
                 field_args = []
                 sd = symbol_map[field.operator]
 
-                field_str, field_args = self._normalize_val_SQL(
+                field_str, field_args = self.render_field_sql(
                     schema,
                     sd,
                     field,
                 )
 
+                if field.or_clause:
+                    if not or_clause:
+                        query_str.append("(")
+                        or_clause = True
+
                 query_str.append('  {}'.format(field_str))
                 query_args.extend(field_args)
 
+                if or_clause:
+                    if not field.or_clause:
+                        query_str.append(")")
+                        or_clause = False
+
+        return query_str, query_args
+
+    def render_sort_sql(self, schema, query, **kwargs):
+        query_str = []
+        query_args = []
+
         if query.fields_sort:
-            query_sort_str = []
             query_str.append('ORDER BY')
+
+            query_sort_str = []
             for field in query.fields_sort:
                 sort_dir_str = 'ASC' if field.direction > 0 else 'DESC'
                 if field.value:
-                    field_sort_str, field_sort_args = self._normalize_sort_SQL(field.name, field.value, sort_dir_str)
+                    field_sort_str, field_sort_args = self.render_sort_field_sql(
+                        field.name,
+                        field.value,
+                        sort_dir_str
+                    )
                     query_sort_str.append(field_sort_str)
                     query_args.extend(field_sort_args)
 
                 else:
                     query_sort_str.append('  {} {}'.format(field.name, sort_dir_str))
 
-            query_str.append(',{}'.format(os.linesep).join(query_sort_str))
+            query_str.append(',\n'.join(query_sort_str))
 
-        if query.bounds:
-            query_str.append(self._normalize_bounds_SQL(query.bounds, sql_options))
+        return query_str, query_args
+
+    def render_bounds_sql(self, schema, query, **kwargs):
+        """
+        https://www.postgresql.org/docs/current/queries-limit.html
+        https://www.sqlite.org/lang_select.html#the_limit_clause
+        """
+        query_str = []
+        query_args = []
+
+        bounds = query.bounds
+        fetchone = kwargs.get("fetchone", False)
+        if bounds or fetchone:
+            if fetchone:
+                limit = 1
+                offset = bounds.offset
+
+            else:
+                if bounds.has_limit():
+                    limit, offset = bounds.get()
+
+                else:
+                    limit = self.LIMIT_NONE
+                    offset = bounds.offset
+
+            query_str.append(f'LIMIT {limit} OFFSET {offset}')
+
+        return query_str, query_args
+
+    def render_sql(self, schema, query, **kwargs):
+        """
+        convert the query instance into SQL
+
+        this is the glue method that translates the generic Query() instance to
+        the SQL specific query, this is where the magic happens
+
+        https://www.sqlite.org/lang_select.html
+
+        :param **kwargs:
+            - count_query: bool, True if this is a count query SELECT
+            - only_where_clause, bool, True to only return after WHERE ...
+        :returns: tuple[str, list[Any]], (query_str, query_args)
+        """
+        query_str = []
+        query_args = []
+
+        select_str, select_args = self.render_select_sql(
+            schema,
+            query,
+            **kwargs
+        )
+        query_str.extend(select_str)
+        query_args.extend(select_args)
+
+        compound_str, compound_args = self.render_compound_sql(
+            schema,
+            query,
+            **kwargs
+        )
+        query_str.extend(compound_str)
+        query_args.extend(compound_args)
+
+        where_str, where_args = self.render_where_sql(
+            schema,
+            query,
+            **kwargs
+        )
+        query_str.extend(where_str)
+        query_args.extend(where_args)
+
+        sort_str, sort_args = self.render_sort_sql(
+            schema,
+            query,
+            **kwargs
+        )
+        query_str.extend(sort_str)
+        query_args.extend(sort_args)
+
+        limit_str, limit_args = self.render_bounds_sql(
+            schema,
+            query,
+            **kwargs
+        )
+        query_str.extend(limit_str)
+        query_args.extend(limit_args)
 
         query_str = "\n".join(query_str)
         return query_str, query_args
@@ -570,12 +778,12 @@ class SQLInterface(SQLInterfaceABC):
         field_names = []
         query_vals = []
         for field_name, field_val in fields.items():
-            field_names.append(self._normalize_name(field_name))
-            field_formats.append(self.val_placeholder)
+            field_names.append(self.render_field_name_sql(field_name))
+            field_formats.append(self.PLACEHOLDER)
             query_vals.append(field_val)
 
         query_str = 'INSERT INTO {} ({}) VALUES ({})'.format(
-            self._normalize_table_name(schema),
+            self.render_table_name_sql(schema),
             ', '.join(field_names),
             ', '.join(field_formats),
         )
@@ -584,7 +792,7 @@ class SQLInterface(SQLInterfaceABC):
             # https://www.sqlite.org/lang_returning.html
             pk_name = schema.pk_name
             if pk_name:
-                query_str += ' RETURNING {}'.format(self._normalize_name(pk_name))
+                query_str += ' RETURNING {}'.format(self.render_field_name_sql(pk_name))
 
         return query_str, query_vals
 
@@ -593,17 +801,17 @@ class SQLInterface(SQLInterfaceABC):
         query_args = []
 
         if not kwargs.get("only_set_clause", False):
-            query_str = 'UPDATE {} '.format(self._normalize_table_name(schema))
+            query_str = 'UPDATE {} '.format(self.render_table_name_sql(schema))
 
         field_str = []
         for field_name, field_val in fields.items():
-            field_str.append('{} = {}'.format(self._normalize_name(field_name), self.val_placeholder))
+            field_str.append('{} = {}'.format(self.render_field_name_sql(field_name), self.PLACEHOLDER))
             query_args.append(field_val)
 
         query_str += 'SET {}'.format(',\n'.join(field_str))
 
         if query:
-            where_query_str, where_query_args = self.get_SQL(schema, query, only_where_clause=True)
+            where_query_str, where_query_args = self.render_sql(schema, query, only_where_clause=True)
             query_str += ' {}'.format(where_query_str)
             query_args.extend(where_query_args)
 
@@ -661,7 +869,7 @@ class SQLInterface(SQLInterfaceABC):
             if field.is_ref():
                 field_type += ' ' + self.render_datatype_ref_sql(field_name, field)
 
-        return '{} {}'.format(self._normalize_name(field_name), field_type)
+        return '{} {}'.format(self.render_field_name_sql(field_name), field_type)
 
     def render_datatype_bool_sql(self, field_name, field, **kwargs):
         return 'BOOL'
@@ -713,8 +921,8 @@ class SQLInterface(SQLInterfaceABC):
             format_str = 'REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE SET NULL'
 
         ret = format_str.format(
-            self._normalize_table_name(ref_s),
-            self._normalize_name(ref_s.pk.name)
+            self.render_table_name_sql(ref_s),
+            self.render_field_name_sql(ref_s.pk.name)
         )
 
         return ret
@@ -728,15 +936,20 @@ class SQLInterface(SQLInterfaceABC):
             placeholders: boolean, True if place holders should remain
         :returns: string if placeholders is False, (string, list) if placeholders is True
         """
-        sql, sql_args = self.get_SQL(schema, query)
+        sql, sql_args = self.render_sql(schema, query)
         placeholders = kwargs.get("placeholders", kwargs.get("placeholder", False))
 
         if not placeholders:
             for sql_arg in sql_args:
-                sa = String(sql_arg)
-                if not sa.isnumeric():
-                    sa = "'{}'".format(sa)
-                sql = sql.replace(self.val_placeholder, sa, 1)
+                if sql_arg is None:
+                    sa = 'NULL'
+
+                else:
+                    sa = String(sql_arg)
+                    if not sa.isnumeric():
+                        sa = "'{}'".format(sa)
+
+                sql = sql.replace(self.PLACEHOLDER, sa, 1)
 
         return (sql, sql_args) if placeholders else sql
 
